"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var SideralObject_1 = require("./../SideralObject");
/**
 * This class is used to generate events and properly destroy it when a SideralObject is killed
 */
var Signal = (function () {
    /* METHODS */
    /**
     * @constructor
     * @param parent - The parent of the signal
     */
    function Signal(parent, onBind, onUnbind) {
        /**
         * List of listeners connected to the signal
         * @private
         */
        this._listeners = [];
        if (!(parent instanceof SideralObject_1.SideralObject)) {
            throw new Error("Signal.constructor: an instance of SideralObject must be passed on the constructor (the owner of the signal).");
        }
        this.parent = parent;
        this.onBind = onBind;
        this.onUnbind = onUnbind;
    }
    /**
     * Dispatch a signal to all listeners with parameters
     * @param params - Parameters
     */
    Signal.prototype.dispatch = function () {
        var params = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            params[_i] = arguments[_i];
        }
        this._listeners.forEach(function (listener) { return listener.apply(void 0, params); });
    };
    /**
     * Add a listener to the current signal
     * @param listener - The listener
     * @returns The listener
     */
    Signal.prototype.add = function (listener) {
        this._listeners.push(listener);
        if (this.onBind) {
            this.onBind(listener);
        }
        return listener;
    };
    /**
     * Remove a listener from the signal
     * @param listener - The listener to remove
     * @returns True if listener has been removed
     */
    Signal.prototype.remove = function (listener) {
        var nextListeners = this._listeners.filter(function (x) { return x !== listener; }), hasRemoved = nextListeners.length !== this._listeners.length;
        this._listeners = nextListeners;
        if (this.onUnbind && hasRemoved) {
            this.onUnbind(listener);
        }
        return hasRemoved;
    };
    /**
     * Remove all listeners from the signal
     */
    Signal.prototype.removeAll = function () {
        this._listeners.forEach(this.remove.bind(this));
        this._listeners = [];
    };
    /**
     * Bind a listener by the first parameter of a trigger
     * @param param - The parameter binded to the listener
     * @param listener - The listener
     * @returns The listener
     */
    Signal.prototype.bind = function (param, listener) {
        var params = [].concat(param);
        return this.add(function (key) {
            if (typeof key === "string" && params.find(function (x) { return x === key; })) {
                listener.apply(void 0, [].slice.call(arguments, 1));
            }
        });
    };
    return Signal;
}());
exports.Signal = Signal;
