"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Tool_1 = require("./../Tool");
var index_1 = require("./../index");
/**
 * Memory let you preload and use resources for your game
 */
var Assets = (function () {
    function Assets() {
    }
    /* METHODS */
    /**
     * Load all assets from an environment
     * @param env - Environnement to load
     * @param afterCallback - Callback to be launched after the full load
     * @param onStepFunction - Function to be called to know the step of loading
     */
    Assets.load = function (env, afterCallback, onStepFunction) {
        env = env || "global";
        var loader = Assets.getLoader(env);
        var pixiLoaded = !Object.keys(loader.pixi.resources).length, soundLoaded = !Object.keys(loader.sound.sounds).length, callbackCalled = false;
        var onLoad = function () {
            if (loader.onLoaded && soundLoaded && pixiLoaded && !callbackCalled) {
                callbackCalled = true;
                loader.onLoaded();
            }
        }, onSoundLoaded = function () {
            soundLoaded = true;
            onLoad();
        }, onPixiLoaded = function () {
            pixiLoaded = true;
            onLoad();
        }, stepFunction = function () {
            if (loader.onStep) {
                loader.onStep(Assets.getLoadProgress(env));
            }
        };
        Assets.env = env;
        loader.loading = true;
        loader.onLoaded = afterCallback || loader.onLoaded;
        loader.onStep = onStepFunction || loader.onStep;
        if (loader.queue.filter(function (queue) { return !queue.loaded; }).length) {
            return null;
        }
        if (soundLoaded && pixiLoaded) {
            onLoad();
            return null;
        }
        loader.pixi.onProgress.add(stepFunction);
        loader.sound.onProgress(stepFunction);
        loader.pixi.load(onPixiLoaded);
        loader.sound.load(onSoundLoaded);
    };
    /**
     * Preload an image
     * @param id - Id of the image
     * @param url - Url of the image to preload
     * @param env - Choose an environment to preload the assets
     * @param options - Options to pass to the pixi Resource-Loader
     */
    Assets.preload = function (id, url, env, options) {
        var loader = Assets.getLoader(env), queue = loader.queue.find(function (x) { return x.id === id; });
        if (!loader.pixi.resources[id]) {
            loader.pixi.add(id, url, options);
        }
        if (queue) {
            queue.loaded = true;
            if (!loader.queue.filter(function (queue) { return !queue.loaded; }).length && loader.loading) {
                Assets.load(env);
            }
        }
        return Assets;
    };
    Assets.preloadBase64 = function (id, base64, env) {
        var loader = Assets.getLoader(env), queue = { id: id, loaded: false };
        loader.queue.push(queue);
        fetch(base64)
            .then(function (res) { return res.blob(); })
            .then(function (blob) { return window.URL.createObjectURL(blob); })
            .then(function (url) { return Assets.preload(id, base64, env, { loadType: index_1.PIXI.loaders.Resource.LOAD_TYPE.IMAGE }); })
            .catch(function () { return loader.queue.filter(function (x) { return x.id !== queue.id; }); });
        return Assets;
    };
    /**
     * Preload a sound
     * @param id - id of the sound
     * @param url - Url of the sound to preload
     * @param env - Choose an environment to preload the assets
     */
    Assets.preloadSound = function (id, url, env) {
        var loader = Assets.getLoader(env);
        loader.sound.preload(id, url);
        return Assets;
    };
    /**
     * Preload all assets for a tilemap
     * @param data - data of a tilemap
     * @param env - Environment to preload the assets
     */
    Assets.preloadTilemap = function (data, env) {
        if (data.backgrounds) {
            data.backgrounds.forEach(function (background) { return Assets.preload(background.path, background.path, env); });
        }
        if (data.decorators) {
            Object.keys(data.decorators.data).forEach(function (key) { return Assets.preload(data.decorators.data[key], data.decorators.data[key], env); });
        }
        Assets.preload(data.path, data.path, env);
        return Assets;
    };
    /**
     * Know if a loader is ready
     * @param env - Environment of the loader
     */
    Assets.isReady = function (env) {
        if (env === void 0) { env = "global"; }
        var loader = Assets.getLoader(env);
        return loader.pixi.progress >= 100 && !loader.pixi.loading && loader.sound.isReady();
    };
    /**
     * Get the PIXI Loader by its environment
     * @param env - Loader of the environment
     */
    Assets.getLoader = function (env) {
        if (env === void 0) { env = "global"; }
        var loader = Assets.Loaders[env];
        if (!loader) {
            loader = Assets.removeLoader(env);
        }
        return loader;
    };
    /**
     * Remove a loader and all its assets
     * @param env - Environment to remove
     */
    Assets.removeLoader = function (env) {
        if (env === void 0) { env = "global"; }
        return Assets.Loaders[env] = {
            pixi: new index_1.PIXI.loaders.Loader(),
            sound: new Tool_1.SoundLoader(),
            queue: [],
            onLoaded: null,
            onStep: null,
            loading: false
        };
    };
    /**
     * Get a resource loaded
     * @param id - Id of the image
     * @param onLoad - Callback when the resource is ready
     * @param env - The environment to get the resource
     */
    Assets.get = function (id, onLoad, env) {
        var loader = Assets.getLoader(env), findResource = function (id) { return loader.pixi.resources[id]; }, resourceExists = function (id) { return loader.pixi.resources[id] || loader.queue.find(function (queue) { return queue.id === id; }); };
        var resources = Array.isArray(id) ? id.map(resourceExists) : resourceExists(id), validated = false;
        if (Array.isArray(resources)) {
            resources = resources.filter(function (resource) { return Boolean(resource); });
            validated = Boolean(resources.length);
        }
        else {
            validated = Boolean(resources);
        }
        if (!validated) {
            console.warn("Assets warning: some image are missing. [" + (Array.isArray(id) ? id.join(",") : id) + "]");
            return null;
        }
        if (Assets.isReady(env)) {
            onLoad(resources);
        }
        else {
            loader.pixi.onComplete.add(function () { return onLoad(Array.isArray(id) ? id.map(findResource) : findResource(id)); });
        }
    };
    /**
     * Get an image loaded
     * @param id - Id of the image
     * @param onLoad - Callback when the resource is ready
     * @param env - The environment to get the resource
     */
    Assets.getImage = function (id, onLoad, env) {
        Assets.get(id, function (resource) { return onLoad(Array.isArray(resource) ? resource.map(function (x) { return x.data; }) : resource.data); }, env);
    };
    /**
     * Get a texture loaded
     * @param id - Id of the image
     * @param onLoad - Callback when the resource is ready
     * @param env - The environment to get the resource
     */
    Assets.getTexture = function (id, onLoad, env) {
        Assets.get(id, function (resource) { return onLoad(Array.isArray(resource) ? resource.map(function (x) { return x.texture; }) : resource.texture); }, env);
    };
    /**
     * Get a Sound Manager
     * @param env - The environment of sound
     * @returns The sound manager
     */
    Assets.getSound = function (env) {
        return this.getLoader(env).sound.manager;
    };
    /**
     * Get the total progression of loading
     * @param env - The environment target
     * @returns The total progression of loading
     */
    Assets.getLoadProgress = function (env) {
        var loader = Assets.getLoader(env), pixiProgress = loader.pixi.progress, howlerProgress = loader.sound.progress;
        return Math.floor((pixiProgress + howlerProgress) / 2);
    };
    return Assets;
}());
/* ATTRIBUTES */
/**
 * PIXI & Sound Loaders
 * @readonly
 */
Assets.Loaders = {};
/**
 * Current environment of the Assets
 * @readonly
 */
Assets.env = "global";
exports.Assets = Assets;
