import { SideralObject, Signal } from "./../SideralObject";
import { Enum } from "./../Tool";
import { IKeyboardSignals } from "./../Interface";


/**
 * Keyboard event utils
 */
export class Keyboard extends SideralObject {

    /* ATTRIBUTES */

    /**
     * Signals emitted by the Keyboard
     * @readonly
     */
    signals: IKeyboardSignals;

    /**
     * list of all inputs key released or pressed relative to the game looping
     * @readonly
     */
    inputs: any = {};

    /**
     * List of all inputs key released or pressed
     * @private
     */
    _inputs: any = {};

    /**
     * If true, the key event willnot be propaged
     */
    preventInputPropagation: boolean = true;


    /* LIFECYCLE */

    /**
     * @constructor
     */
    constructor () {
        super();

        this.signals.keyChange  = new Signal(this);
        this.signals.keyPress   = new Signal(this);
        this.signals.keyRelease = new Signal(this);

        this.connect(this.signals.update, this._updateInputs.bind(this));
    }

    /**
     * @override
     */
    initialize (props = {}): void {
        super.initialize(props);

        window.addEventListener("keydown", this._onKeydown.bind(this));
        window.addEventListener("keyup", this._onKeyup.bind(this));
    }

    /**
     * @override
     */
    kill (): void {
        super.kill();

        window.removeEventListener("keydown", this._onKeydown.bind(this));
        window.removeEventListener("keyup", this._onKeydown.bind(this));
    }


    /* EVENTS */

    /**
     * Update all device inputs
     * @private
     */
    _updateInputs (): void {
        for (const key in this._inputs) {
            if (!this._inputs.hasOwnProperty(key)) {
                continue;
            }

            const input = this.inputs[key],
                _input = this._inputs[key];

            // Pressed
            if (_input === Enum.KEY_STATE.PRESSED) {
                if (input === _input) {
                    this.inputs[key] = Enum.KEY_STATE.HOLD;

                } else if (input !== Enum.KEY_STATE.HOLD) {
                    this.inputs[key] = Enum.KEY_STATE.PRESSED;
                    this.signals.keyChange.dispatch(key, true);
                    this.signals.keyPress.dispatch(key);
                }

                // Released
            } else if (_input === Enum.KEY_STATE.RELEASED) {
                if (!input) {
                    this.inputs[key] = Enum.KEY_STATE.PRESSED;

                } else if (input === _input) {
                    delete this.inputs[key];
                    delete this._inputs[key];

                } else {
                    this.inputs[key] = Enum.KEY_STATE.RELEASED;
                    this.signals.keyChange.dispatch(key, false);
                    this.signals.keyRelease.dispatch(key);
                }
            }
        }
    }

    /**
     * Event on keydown
     * @param e - The event
     * @returns Input propagation
     */
    _onKeydown (e: KeyboardEvent): boolean {
        if (this.preventInputPropagation) {
            e.preventDefault();
            e.stopPropagation();
        }

        this._inputs[e.keyCode] = Enum.KEY_STATE.PRESSED;

        return !this.preventInputPropagation;
    }

    /**
     * Event on keyup
     * @param e - The event
     * @returns Input propagation
     */
    _onKeyup (e: KeyboardEvent): boolean {
        if (this.preventInputPropagation) {
            e.preventDefault();
            e.stopPropagation();
        }

        this._inputs[e.keyCode] = Enum.KEY_STATE.RELEASED;

        return !this.preventInputPropagation;
    }
}
