import { Signal } from "./../SideralObject";
import { IProps, ISlot, ISignals, IAddMultiple, IContext } from "./../Interface";


/**
 * The entry class of all object in Sideral
 */
export class SideralObject {

    /* ATTRIBUTES */

    /**
     * Unique id of the object
     * @readonly
     */
    id: number = SideralObject.generateId();

    /**
     * Name of the object (used to identify object, you should not forget to fill this value)
     * @type {string}
     */
    name: string = "";

    /**
     * Properties of the object
     */
    props: IProps = {};

    /**
     * Last values of properties of the object
     */
    last: any = {};

    /**
     * Context is a object which you can store anything, the content of the context will be passed to its children
     */
    context: IContext = {};

    /**
     * List of all signals of the element
     */
    signals: ISignals = {
        update: new Signal(this),
        propChange: new Signal(this),
        addChild: new Signal(this),
        removeChild: new Signal(this)
    };

    /**
     * List of all children of this object
     */
    children: SideralObject[] = [];

    /**
     * Parent of this object
     */
    parent: SideralObject = null;

    /**
     * Know if this object has been initialized
     * @readonly
     */
    initialized: boolean = false;

    /**
     * Know if this object has been killed
     * @readonly
     */
    killed: boolean = false;

    /**
     * List of all slots connected to signals
     * @private
     */
    _slots: Array<ISlot> = [];


    /* LIFECYCLE */

    /**
     * Lifecycle - When initialized by a parent (called only once when the instance is attached to the lifecycle of the game)
     * @access public
     * @param props - properties to merge
     */
    initialize (props: any = {}): void {
        this.setProps(props);
        this.initialized = true;
    }

    /**
     * Lifecycle - Destroy the current instance
     * @access public
     */
    kill (): void {
        Object.keys(this.signals).forEach(key => {
            const signal = this.signals[key];

            signal.removeAll();
            signal.disabled = true;
        });

        this._slots.forEach(slot => slot.listeners.forEach(listener => slot.signal.remove(listener)));
        this.children.forEach(child => child.kill());

        this.killed     = true;
        this._slots     = [];
        this.children   = [];

        if (this.parent) {
            this.parent.children = this.parent.children.filter(child => child.id !== this.id);
            this.parent.signals.removeChild.dispatch(this.name, this);
            this.parent = null;
        }
    }

    /**
     * Lifecycle - Called every loop
     * @access protected
     * @param {number} tick - The tick factor (to prevent the dependance of the framerate)
     * @returns {void}
     */
    update (tick: number): void {
        this.children.forEach(child => !child.killed && child.update(tick));
        this.signals.update.dispatch(tick);
    }

    /**
     * Lifecycle - Called before a next game loop
     * @access protected
     * @returns {void}
     */
    nextCycle (): void {
        const propChanged = [];

        this.children.forEach(child => child.nextCycle());

        Object.keys(this.props).forEach(key => {
            const prop = this.props[key];

            if (typeof prop !== "object" && prop !== this.last[key]) {
                propChanged.push(key);
                this.last[key] = this.props[key];

            }
        });

        propChanged.forEach(prop => this.signals.propChange.dispatch(prop, this.props[prop]));
    }


    /* METHODS */

    /**
     * Set new properties to the object. All attribute contained in "props" are public and can be edited by external source.
     * Properties can be observe via the "propChange" event. The "propChange" signal will not be triggered if the property is created.
     * @access public
     * @param props - Properties to merge
     * @param force - If true, the properties will be forced and "propChange" will not be fired
     * @returns Current instance
     * @example
     *  this.setProps({
     *      test: 1
     *  });
     *
     *  this.props.test; // 1
     */
    setProps (props: any = {}, force?: boolean): this {
        Object.keys(props).forEach(key => {
            const prop = this.props[key] = props[key];

            if (typeof this.last[key] === "undefined" || force) {
                this.last[key] = prop;
            }
        });

        return this;
    }

    /**
     * Remove a property from the lifecycle of the object
     * @param name - Name of the property to remove
     * @returns Current instance
     */
    removeProp (name: string): this {
        delete this.last[name];
        delete this.props[name];

        return this;
    }

    /**
     * Add an item to the current object. The item added will enter into the lifecycle of the object and will become a children
     * of this object. The method "initialize" of the item will be called.
     * @access public
     * @param item - A SideralObject
     * @param props - Props to merge to the item
     * @returns The item initialized
     */
    add (item: SideralObject, props: any = {}): SideralObject {
        if (!(item instanceof SideralObject)) {
            throw new Error("SideralObject.add : item must be an instance of Sideral Abstract Class");
        }

        item.parent = this;

        Object.keys(this.context).forEach(key => item.context[key] = this.context[key]);
        this.children.push(item);
        item.initialize(props);
        this.signals.addChild.dispatch(item.name, item);

        return item;
    }

    /**
     * Check if a property (an attribute from "this.props") has changed
     * @access public
     * @param propName - name of the property to check
     * @returns Property has changed ?
     */
    hasChanged (propName: string): boolean {
        if (!this.props[propName]) {
            return false;
        }

        return this.props[propName] !== this.last[propName];
    }

    /**
     * Connect a listener to a signal (automatically removed when this object is killed)
     * You can bind a parameter to the listener. In that case, the listerner will be trigger only if the first parameter
     * of the dispatch is corresponding to the parameter
     * 
     * @example
     *  this.connect(this.sideralObject.signals.propChange, "value", () => console.log(true));
     *  this.sideralObject.signals.propChange.dispatch("toto", true); // the listener above will not be triggered
     *  this.sideralObject.signals.propChange.dispatch("value", 4); // the listener above will now be triggered
     *  
     * @param signal - The signal
     * @param listenerOrParam - The listener or the param to bind to the listener
     * @param listeners - The rest of listeners
     */
    connect (signal: Signal, listenerOrParam: string | Array<string> | Function, ...listeners: Array<Function>): this {
        if (!(signal instanceof Signal)) {
            throw new Error("SideralObject.connect: the first argument 'signal' must be an instance of Signal.");
        }

        const slotExists: ISlot = this._slots.find(x => x.signal === signal);
        const slot: ISlot       = slotExists || {
            signal: signal,
            listeners: []
        };

        if (typeof listenerOrParam === "function") {
            [].concat(listenerOrParam, listeners).forEach(listener => slot.listeners.push(signal.add(listener)));

        } else if (((typeof listenerOrParam === "string") || Array.isArray(listenerOrParam))) {
            listeners.forEach(listener => slot.listeners.push(signal.bind(listenerOrParam, listener)));

        }

        if (!slotExists) {
            this._slots.push(slot);
        }

        return this;
    }


    /* STATICS */

    /**
     * Generate an unique id
     * @returns The unique id
     */
    static generateId(): number {
        return Math.floor((1 + Math.random()) * 0x100000);
    }
}
